/*
    Copyright 2007-2012 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>
#include <windows.h>
#include <tlhelp32.h>
#include "quickrva_res.h"
#include "distorm.h"

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;
typedef uint64_t    u64;



#define VER2        VER "\n" \
                    "by Luigi Auriemma\n" \
                    "e-mail: aluigi@autistici.org\n" \
                    "web:    aluigi.org\n" \
                    "  disassembler engine from diStorm64\n" \
                    "\n" \
                    "quick tool for exploring PE and ELF files\n" \
                    "\n" \
                    "Usage tips:\n" \
                    "the case-sensitive search function supports C style escape\n" \
                    "sequences (like \\n, \\r, \\x00 and so on) so you can search\n" \
                    "really any type of char and binary pattern in the file\n" \
                    "Use \"prev\" for the previous occurrencies\n" \
                    "and \"next\" for the subsequent occurrencies\n" \
                    "\n" \
                    "The tool automatically understands if the input is a RVA\n" \
                    "or file address and displays the data at that offset and\n" \
                    "the informations about the section which contains it\n" \
                    "\n" \
                    "When you load a process the \"file\" field will contain the offset\n" \
                    "in memory, not that on its original executable file because it's\n" \
                    "everything referred to the file mapped in memory\n" \
                    "\n" \
                    "Drag'n'drop, command-line file loading and process loading supported\n" \
                    "The tool supports also the navigation of the hex dump window with\n" \
                    "keys and intuitive keyboard shortcuts like CTRL-F for search, F3 for\n" \
                    "next occurrency, CTRL-O and CTRL-P for opening files and processes\n" \
                    "and so on"
#define MAXLINES    8
#define MAXLINELEN  68
#define MAXPAGE     128     // MAXLINES * 16
#define SEARCHSZ    128
#define TEXTSZ      1024
#define PROCFMT     "%10u %08x %08x %s"
#define SECNAMESZ   32
#define MYPAD(X)    ((X + (sec_align - 1)) & (~(sec_align - 1)))
#define MYALIGNMENT 0x1000  // default in case not available



typedef struct {
	DWORD Signature;
	IMAGE_FILE_HEADER FileHeader;
	IMAGE_OPTIONAL_HEADER32 OptionalHeader;
} MYIMAGE_NT_HEADERS32;

typedef struct {
	DWORD Signature;
	IMAGE_FILE_HEADER FileHeader;
	IMAGE_OPTIONAL_HEADER64 OptionalHeader;
} MYIMAGE_NT_HEADERS64;

typedef struct {    // from http://hte.sf.net
    u32     vsize;
    u32     base_reloc_addr;
    u32     flags;
    u32     page_map_index;
    u32     page_map_count;
    u8      name[4];
} vxd_section_t;

typedef struct {
    u8      e_ident[16];
    u16     e_type;
    u16     e_machine;
    u32     e_version;
    u32     e_entry;
    u32     e_phoff;
    u32     e_shoff;
    u32     e_flags;
    u16     e_ehsize;
    u16     e_phentsize;
    u16     e_phnum;
    u16     e_shentsize;
    u16     e_shnum;
    u16     e_shstrndx;
} elf32_header_t;

typedef struct {
    u8      e_ident[16];
    u16     e_type;
    u16     e_machine;
    u32     e_version;
    u64     e_entry;
    u64     e_phoff;
    u64     e_shoff;
    u32     e_flags;
    u16     e_ehsize;
    u16     e_phentsize;
    u16     e_phnum;
    u16     e_shentsize;
    u16     e_shnum;
    u16     e_shstrndx;
} elf64_header_t;

typedef struct {
    u32     sh_name;
    u32     sh_type;
    u32     sh_flags;
    u32     sh_addr;     
    u32     sh_offset;
    u32     sh_size;
    u32     sh_link;
    u32     sh_info;
    u32     sh_addralign;
    u32     sh_entsize;
} elf32_section_t;

typedef struct {
    u32     sh_name;
    u32     sh_type;
    u64     sh_flags;
    u64     sh_addr;
    u64     sh_offset;
    u64     sh_size;
    u32     sh_link;
    u32     sh_info;
    u64     sh_addralign;
    u64     sh_entsize;
} elf64_section_t;

typedef struct {    // from http://hte.sf.net
    u32     magic_id;
    u8      signature[256];
    u32     base_address;
    u32     size_of_headers;
    u32     size_of_image;
    u32     size_of_imageheader;
    u32     timedate;
    u32     certificate_address;
    u32     number_of_sections;
    u32     section_header_address;
    u32     initialisation_flags;
    u32     entry_point;
    u32     tls_address;
    u32     pe_stack_commit;
    u32     pe_heap_reserve;
    u32     pe_heap_commit;
    u32     pe_base_address;
    u32     pe_size_of_image;
    u32     pe_checksum;
    u32     pe_timedate;
    u32     debug_pathname_address;
    u32     debug_filename_address;
    u32     debug_unicode_filename_address;
    u32     kernel_image_thunk_address;
    u32     non_kernel_import_directory_address;
    u32     number_of_library_versions;
    u32     library_versions_address;
    u32     kernel_library_version_address;
    u32     xapi_library_version_address;
    u32     logo_bitmap_address;
    u32     logo_bitmap_size;
} xbe_header_t;

typedef struct {    // from http://hte.sf.net
    u32     section_flags;
    u32     virtual_address;
    u32     virtual_size;
    u32     raw_address;
    u32     raw_size;
    u32     section_name_address;
    u32     section_name_ref_count;
    u32     head_shared_page_ref_count_address;
    u32     tail_shared_page_ref_count_address;
    u8      section_digest[20];
} xbe_section_t;

typedef struct {
    u8      Name[SECNAMESZ + 1];
    u32     VirtualAddress;
    u32     VirtualSize;
    u32     PointerToRawData;
    u32     SizeOfRawData;
    u32     Characteristics;
} section_t;



section_t   *section            = NULL;
_DecodeType fileasmsz           = Decode32Bits;
u32     filememsz               = 0,
        curroff                 = -1,
        imagebase               = 0,
        sec_align               = 0,
        entrypoint              = 0;
int     isprocess               = 0,
        sections                = 0;
u8      filename[TEXTSZ + 32]   = "",
        *filemem                = NULL,
        *commandline;



void show_dump(unsigned char *data, unsigned int len, unsigned char *out) {
    static const u8 hex[16] = "0123456789ABCDEF";
    u8      buff[67],
            chr,
            *bytes,
            *p,
            *o,
            *limit,
            *glimit;

    if(!len) {
        out[0] = 0;
        return;
    }
    o = out;

    glimit = data + len;
    memset(buff + 2, ' ', 48);

    while(data < glimit) {
        limit = data + 16;
        if(limit > glimit) {
            limit = glimit;
            memset(buff, ' ', 48);
        }

        p     = buff;
        bytes = p + 50;
        while(data < limit) {
            chr = *data;
            *p++ = hex[chr >> 4];
            *p++ = hex[chr & 15];
            p++;
            *bytes++ = ((chr < ' ') || (chr >= 0x7f)) ? '.' : chr;
            data++;
        }

        memcpy(o, buff, bytes - buff);
        o += bytes - buff;
        *o++ = '\r';
        *o++ = '\n';
    }
    o[-2] = 0;
}



void process_list(HWND hwnd) {
    PROCESSENTRY32  Process;
    MODULEENTRY32   Module;
    HANDLE          snapProcess,
                    snapModule;
    BOOL            b;
    u8              process_name[TEXTSZ];

#define STARTPROC(X,Y) \
            snap##X = CreateToolhelp32Snapshot(Y, Process.th32ProcessID); \
            X.dwSize = sizeof(X); \
            for(b = X##32First(snap##X, &X); b; b = X##32Next(snap##X, &X)) { \
                X.dwSize = sizeof(X);
#define ENDPROC(X) \
            } \
            CloseHandle(snap##X);

    SendDlgItemMessage(hwnd, IDC_PROCLIST, LB_ADDSTRING, 0, (LPARAM)
        "       pid baseaddr basesize process_path");
    SendDlgItemMessage(hwnd, IDC_PROCLIST, LB_ADDSTRING, 0, (LPARAM)
        "-----------------------------------------");

    Process.th32ProcessID = 0;
    STARTPROC(Process, TH32CS_SNAPPROCESS)
        STARTPROC(Module, TH32CS_SNAPMODULE)

            snprintf(process_name, sizeof(process_name),
                PROCFMT,
                (u32)Module.th32ProcessID,
                (u32)Module.modBaseAddr,
                (u32)Module.modBaseSize,
                Module.szExePath);
            process_name[sizeof(process_name) - 1] = 0;

            SendDlgItemMessage(hwnd, IDC_PROCLIST, LB_ADDSTRING, 0, (LPARAM)process_name);

            break;  // we need only the first one!

        ENDPROC(Module)
    ENDPROC(Process)

#undef STARTPROC
#undef ENDPROC
}



u8 *itox(u32 num) {
    static u8   ret[9];

    sprintf(ret, "%08x", num);
    return(ret);
}



void global_reset(HWND hwnd) {
    fileasmsz   = Decode32Bits;
    filememsz   = 0;
    imagebase   = 0;
    sec_align   = 0;
    entrypoint  = 0;
    isprocess   = 0;
    if(filemem) free(filemem);
    filemem     = NULL;
    if(section) free(section);
    section     = NULL;
    sections    = 0;
    curroff     = -1;
    SetDlgItemText(hwnd, IDC_FILENAME,  "");
    SetDlgItemText(hwnd, IDC_FILESIZE,  "");
    SetDlgItemText(hwnd, IDC_RVA,       "");
    SetDlgItemText(hwnd, IDC_FILE,      "");
    SetDlgItemText(hwnd, IDC_BYTES,     "");
    SetDlgItemText(hwnd, IDC_DISASM,    "");
    SetDlgItemText(hwnd, IDC_ENTRYPOINT,"");
    SetDlgItemText(hwnd, IDC_IMAGEBASE, "");
    SetDlgItemText(hwnd, IDC_SECALIGN,  "");
    SetDlgItemText(hwnd, IDC_NAME,      "");
    SetDlgItemText(hwnd, IDC_VOFFSET,   "");
    SetDlgItemText(hwnd, IDC_VSIZE,     "");
    SetDlgItemText(hwnd, IDC_RAWOFF,    "");
    SetDlgItemText(hwnd, IDC_RAWSIZE,   "");
    SetDlgItemText(hwnd, IDC_CHARAC,    "");
}



void std_err(HWND hwnd) {
    MessageBox(hwnd, strerror(errno), VER, MB_ICONERROR | MB_OK | MB_TASKMODAL);
}



void winerr(HWND hwnd) {
    u8     *message = NULL;

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        GetLastError(),
        0,
        (char *)&message,
        0,
        NULL);
    if(message) {
        MessageBox(hwnd, message, VER, MB_ICONERROR | MB_OK | MB_TASKMODAL);
        LocalFree(message);
    }
}



void set_section(HWND hwnd, int idx) {
    if((idx < 0) || (idx > sections)) {
        SetDlgItemText(hwnd, IDC_NAME,      "");
        SetDlgItemText(hwnd, IDC_VOFFSET,   "");
        SetDlgItemText(hwnd, IDC_VSIZE,     "");
        SetDlgItemText(hwnd, IDC_RAWOFF,    "");
        SetDlgItemText(hwnd, IDC_RAWSIZE,   "");
        SetDlgItemText(hwnd, IDC_CHARAC,    "");
        return;
    }
    SetDlgItemText(hwnd, IDC_NAME,      section[idx].Name);
    SetDlgItemText(hwnd, IDC_VOFFSET,   itox(section[idx].VirtualAddress));
    SetDlgItemText(hwnd, IDC_VSIZE,     itox(section[idx].VirtualSize));
    SetDlgItemText(hwnd, IDC_RAWOFF,    itox(section[idx].PointerToRawData));
    SetDlgItemText(hwnd, IDC_RAWSIZE,   itox(section[idx].SizeOfRawData));
    SetDlgItemText(hwnd, IDC_CHARAC,    itox(section[idx].Characteristics));
}



u32 rva2file(HWND hwnd, u32 va) {
    u32     diff;
    int     i,
            ret;

    va  -= imagebase;
    ret  = -1;
    diff = -1;
    for(i = 0; i < sections; i++) {
        if((sections > 1) && !section[i].VirtualAddress) continue;
        if((va >= section[i].VirtualAddress) && (va < (section[i].VirtualAddress + section[i].VirtualSize))) {
            if((va - section[i].VirtualAddress) < diff) {
                diff = va - section[i].VirtualAddress;
                ret  = i;
            }
        }
    }
    if(hwnd) set_section(hwnd, ret);
    if(ret < 0) return(-1);
    return(section[ret].PointerToRawData + va - section[ret].VirtualAddress);
}



u32 file2rva(HWND hwnd, u32 file) {
    u32     diff;
    int     i,
            ret;

    ret  = -1;
    diff = -1;
    for(i = 0; i < sections; i++) {
        if((file >= section[i].PointerToRawData) && (file < (section[i].PointerToRawData + section[i].SizeOfRawData))) {
            if((file - section[i].PointerToRawData) < diff) {
                diff = file - section[i].PointerToRawData;
                ret  = i;
            }
        }
    }
    if(hwnd) set_section(hwnd, ret);
    if(ret < 0) return(-1);
    return(section[ret].VirtualAddress + file - section[ret].PointerToRawData);
}



int get_section(u32 file) {
    u32     diff;
    int     i,
            ret;

    ret  = -1;
    diff = -1;
    for(i = 0; i < sections; i++) {
        if((file >= section[i].PointerToRawData) && (file < (section[i].PointerToRawData + section[i].SizeOfRawData))) {
            if((file - section[i].PointerToRawData) < diff) {
                diff = file - section[i].PointerToRawData;
                ret  = i;
            }
        }
    }
    return(ret);
}



u8 *strings_filter(int offset) {
    static u8   *ubuff = NULL;
    int     i;
    u8      *string,
            *p,
            c;

    string = filemem + offset;

    if(strlen(string) == 1) {
        if(!ubuff) {
            ubuff = malloc(MAXLINELEN + 1);
            if(!ubuff) return(NULL);
        }

        p = string;
        for(i = 0; i < MAXLINELEN; i++) {
            if(!p[0]) break;
            if(p[1]) break;
            ubuff[i] = *p;
            p += 2;
        }
        ubuff[i] = 0;
        string = ubuff;
    }

    if(strlen(string) < 4) return(NULL);  // at least 4 chars

    for(i = 0; (c = string[i]); i++) {
        if(!((c == '\t') || (c == '\r') || (c == '\n') || ((c >= ' ') && (c < 0x7f)))) return(NULL);
    }
    return(string);
}



u8 *string2c(u8 *string) {
    static u8   *ret = NULL;
    int     i;
    u8      c,
            x,
            *p;

    if(!ret) {
        ret = malloc(1 + (MAXLINELEN * 6) + 1 + 1); // for hex
        if(!ret) return("");
    }

    p = ret;
    p += sprintf(p, "                 ");
    *p++ = '\"';
    for(i = 0; (c = string[i]); i++) {
        if((p - ret) >= (MAXLINELEN - 1)) break;
        switch(c) {
            case '\\': { *p++ = '\\'; *p++ = '\\';  break; }
            case '\"': { *p++ = '\\'; *p++ = '\"';  break; }
            //case '\'': { *p++ = '\\'; *p++ = '\'';  break; }
            //case '\?': { *p++ = '\\'; *p++ = '\?';  break; }
            case '\r': { *p++ = '\\'; *p++ = 'r';   break; }
            case '\n': { *p++ = '\\'; *p++ = 'n';   break; }
            case '\a': { *p++ = '\\'; *p++ = 'a';   break; }
            case '\b': { *p++ = '\\'; *p++ = 'b';   break; }
            case '\f': { *p++ = '\\'; *p++ = 'f';   break; }
            case '\t': { *p++ = '\\'; *p++ = 't';   break; }
            case '\v': { *p++ = '\\'; *p++ = 'v';   break; }
            default:   {
                if(!isprint(c)) {
                    p += sprintf(p, "\\x%02x", c);
                    x = tolower(string[i + 1]);
                    if(((x >= '0') && (x <= '9')) || ((x >= 'a') && (x <= 'f'))) {
                        *p++ = '\"';
                        *p++ = '\"';
                    }
                } else {
                    *p++ = c;
                }
                break;
            }
        }
    }
    if((p - ret) >= (MAXLINELEN - 1)) {
        strcpy(ret + MAXLINELEN - 4, "...");
    } else {
        *p++ = '\"';
        *p = 0;
    }
    return(ret);
}



void dumpasm(HWND hwnd, int advance);   // prototype, needed



int set_offset(HWND hwnd, u32 offset, int dlg) {
    u32     rva,
            file,
            dumpsz;
    u8      dump[(MAXLINES * 80) + 1];

    if(offset == -1) return(-1);
    if(!filemem) return(-1);
    if(dlg == IDC_RVA) {
        if(offset < imagebase) {
            set_offset(hwnd, 0, IDC_FILE);
            return(-1);
        }
        rva = offset;
        if(isprocess) {
            file = rva - imagebase;
        } else {
            file = rva2file(hwnd, rva);
        }
    } else {
        file = offset;
        if(isprocess) {
            rva = file + imagebase;
        } else {
            rva = file2rva(hwnd, file);
            if(rva == -1) {
                rva = imagebase;
            } else {
                rva += imagebase;
            }
        }
    }
    if((file == curroff) || (rva == -1)) return(-1);
    if(file >= filememsz) {
        set_offset(hwnd, filememsz - 1, IDC_FILE);
        return(-1);
    }
    curroff = file;

    dumpsz = MAXPAGE;
    if(dumpsz > (filememsz - file)) dumpsz = filememsz - file;
    show_dump(filemem + file, dumpsz, dump);

    SetDlgItemText(hwnd, IDC_RVA,    itox(rva));
    SetDlgItemText(hwnd, IDC_FILE,   itox(file));
    SetDlgItemText(hwnd, IDC_BYTES,  dump);
    dumpasm(hwnd, 0);
    return(0);
}



int Disasm(u8 *src, u32 srcsize, u32 srcip, u8 **result) {
    static u8       disasm_string[256];
    _DecodeResult   res;
    _DecodedInst    decodedInstructions[15];    // 15 is the minimum!
    int             decodedInstructionsCount;

    res = distorm_decode(srcip, src, srcsize, fileasmsz, decodedInstructions, 15, &decodedInstructionsCount);
    if((res == DECRES_INPUTERR) || (decodedInstructionsCount < 1) || (decodedInstructions[0].size < 1)) {
        if(result) *result = "???";
        return(1);
    }
    if(result) {
        sprintf(disasm_string,
            "%s\t%s%s",
            decodedInstructions[0].mnemonic.p,
            decodedInstructions[0].operands.length ? " " : "",
            decodedInstructions[0].operands.p);
        *result = disasm_string;
    }
    return(decodedInstructions[0].size);
}



u32 Disasmback(u8 *src, u32 srcsize, u32 srcip, u32 limit) {    // not that bad, works almost ever
    u32     lastip;
    int     i,
            tmp,
            back,
            asmsz,
            maxlines;

    maxlines = MAXLINES;
    lastip = srcip;
    back = maxlines;
    for(;;) {
        if((srcip - back) < limit) {
            maxlines--;
            if(!maxlines) return(limit);
            back = maxlines;
            continue;
        }
        tmp = -back;
        for(i = 0; i <= maxlines; i++) {
            asmsz = Disasm(src + tmp, srcsize, srcip + tmp, NULL);
            if(!asmsz) break;
            tmp += asmsz;
            if(tmp > 0) break;
            if(i == (maxlines - 1)) lastip = srcip + tmp;
        }
        if(i > maxlines) break;
        back++;
    }
    return(lastip);
}



void dumpasm(HWND hwnd, int advance) {
    u32     eip,
            file,
            string_offset;
    int     i,
            len,
            idx,
            asm_size;
    u8      dump[MAXLINES * (MAXLINELEN + 1)],
            *limit,
            *code,
            *string,
            *da_result;

    file = curroff;
    dump[0] = 0;

    eip = file2rva(NULL, file);
    if(eip == -1) goto quit;
    eip += imagebase;

    idx = get_section(file);
    if(idx < 0) goto quit;

    code  = filemem + file;
    limit = filemem + section[idx].PointerToRawData + section[idx].SizeOfRawData;

    if(advance) {
        if(advance < 0) {
            advance = -advance;
            for(i = 0; i < advance; i++) {
                eip = Disasmback(code, limit - code, eip, imagebase + section[idx].VirtualAddress);
            }
        } else {
            for(i = 0; i < advance; i++) {
                asm_size = Disasm(code, limit - code, eip, NULL);
                code += asm_size;
                eip  += asm_size;
                if(code >= limit) break;
            }
        }
        set_offset(hwnd, rva2file(hwnd, eip), IDC_FILE);
        return;
    }

    len = 0;
    for(i = 0; i < MAXLINES; i++) {
        asm_size = Disasm(code, limit - code, eip, &da_result);

        string = strstr(da_result, "0x");   // unfortunately distorm doesn't return the offsets as in the previous library I used
        if(string) {
            sscanf(string + 2, "%x", &string_offset);
            string_offset = rva2file(NULL, string_offset);
            if((string_offset != -1) && (string_offset < filememsz)) {
                string = strings_filter(string_offset);
                if(string) {
                    len += sprintf(dump + len, "%s\r\n", string2c(string));
                    i++;
                }
            }
        }
        len += sprintf(dump + len, "%08X %s\r\n", eip, da_result);

        code += asm_size;
        eip  += asm_size;
        if(code >= limit) break;
    }
quit:
    SetDlgItemText(hwnd, IDC_DISASM, dump);
}



int parse_PE(HWND hwnd) {
    IMAGE_DOS_HEADER        *doshdr;
    MYIMAGE_NT_HEADERS32    *nt32hdr;
    MYIMAGE_NT_HEADERS64    *nt64hdr;
    IMAGE_ROM_HEADERS       *romhdr;
    IMAGE_OS2_HEADER        *os2hdr;
    IMAGE_VXD_HEADER        *vxdhdr;
    IMAGE_SECTION_HEADER    *sechdr;
    vxd_section_t           *vxdsechdr;
    u32     tmp;
    int     i;
    u8      *p;

    p = filemem;
    doshdr  = (IMAGE_DOS_HEADER *)p;
    if(doshdr->e_magic != IMAGE_DOS_SIGNATURE) return(-1);

    if((doshdr->e_lfanew >= sizeof(IMAGE_DOS_HEADER)) && doshdr->e_cs) {  // note that the following instructions have been tested on various executables but I'm not sure if they are perfect
        tmp = doshdr->e_cparhdr * 16;
        if(doshdr->e_cs < 0x8000) tmp += doshdr->e_cs * 16;
        p += tmp;
    } else {
        if(doshdr->e_lfanew && (doshdr->e_lfanew < filememsz)) {
            p += doshdr->e_lfanew;
        } else {
            p += sizeof(IMAGE_DOS_HEADER);
        }
    }

    nt32hdr = (MYIMAGE_NT_HEADERS32 *)p;
    nt64hdr = (MYIMAGE_NT_HEADERS64 *)p;
    romhdr  = (IMAGE_ROM_HEADERS *)p;
    os2hdr  = (IMAGE_OS2_HEADER *)p;
    vxdhdr  = (IMAGE_VXD_HEADER *)p;

    if(nt32hdr->Signature == IMAGE_NT_SIGNATURE) {
        if(nt32hdr->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC) {
            fileasmsz   = Decode32Bits;
            p += sizeof(MYIMAGE_NT_HEADERS32);
            imagebase   = nt32hdr->OptionalHeader.ImageBase;
            sec_align   = nt32hdr->OptionalHeader.SectionAlignment;
            entrypoint  = imagebase + nt32hdr->OptionalHeader.AddressOfEntryPoint;
            sections    = nt32hdr->FileHeader.NumberOfSections;
        } else if(nt64hdr->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC) {
            fileasmsz   = Decode64Bits;
            p += sizeof(MYIMAGE_NT_HEADERS64);
            imagebase   = nt64hdr->OptionalHeader.ImageBase;
            sec_align   = nt64hdr->OptionalHeader.SectionAlignment;
            entrypoint  = imagebase + nt64hdr->OptionalHeader.AddressOfEntryPoint;
            sections    = nt64hdr->FileHeader.NumberOfSections;
        } else if(romhdr->OptionalHeader.Magic == IMAGE_ROM_OPTIONAL_HDR_MAGIC) {
            fileasmsz   = Decode32Bits;
            p += sizeof(IMAGE_ROM_HEADERS);
            imagebase   = 0;
            sec_align   = MYALIGNMENT;
            entrypoint  = imagebase + romhdr->OptionalHeader.AddressOfEntryPoint;
            sections    = 0;
            section     = NULL;
            return(0);
        } else {
            return(-1);
        }

        section = calloc(sizeof(section_t), sections);
        if(!section) std_err(hwnd);

        sechdr = (IMAGE_SECTION_HEADER *)p;
        for(i = 0; i < sections; i++) {
            strncpy(section[i].Name, sechdr[i].Name, IMAGE_SIZEOF_SHORT_NAME);
            section[i].VirtualAddress   = sechdr[i].VirtualAddress;
            section[i].VirtualSize      = sechdr[i].Misc.VirtualSize;
            section[i].PointerToRawData = sechdr[i].PointerToRawData;
            section[i].SizeOfRawData    = sechdr[i].SizeOfRawData;
            section[i].Characteristics  = sechdr[i].Characteristics;
            if(!section[i].VirtualSize) section[i].VirtualSize = section[i].SizeOfRawData;  // Watcom
        }

    } else if(os2hdr->ne_magic == IMAGE_OS2_SIGNATURE) {
        fileasmsz   = Decode16Bits;
        p += sizeof(IMAGE_OS2_HEADER);
        imagebase   = 0;
        sec_align   = os2hdr->ne_align;
        entrypoint  = imagebase + os2hdr->ne_csip;
        sections    = 0;
        sechdr      = NULL;

    } else if(
      (vxdhdr->e32_magic == IMAGE_OS2_SIGNATURE_LE) ||  // IMAGE_VXD_SIGNATURE is the same signature
      (vxdhdr->e32_magic == 0x3357) ||                  // LX, W3 and W4: I guess they are the same... I hope
      (vxdhdr->e32_magic == 0x3457) ||
      (vxdhdr->e32_magic == 0x584C)) {
        fileasmsz   = Decode16Bits;
        p += sizeof(IMAGE_VXD_HEADER);
        imagebase   = 0;
        sec_align   = vxdhdr->e32_pagesize;
        entrypoint  = 0;    // handled later
        sections    = vxdhdr->e32_objcnt;

        section = calloc(sizeof(section_t), sections);
        if(!section) std_err(hwnd);

        tmp = vxdhdr->e32_datapage;
        vxdsechdr = (vxd_section_t *)p;
        for(i = 0; i < sections; i++) {
            strncpy(section[i].Name, vxdsechdr[i].name, 4);
            section[i].VirtualAddress   = vxdsechdr[i].base_reloc_addr;
            section[i].VirtualSize      = vxdsechdr[i].vsize;
            section[i].PointerToRawData = tmp;
            section[i].SizeOfRawData    = vxdsechdr[i].vsize;
            section[i].Characteristics  = vxdsechdr[i].flags;
            tmp += MYPAD(section[i].SizeOfRawData);
            if(!entrypoint && (tmp > vxdhdr->e32_eip)) {    // I'm not totally sure if this is correct but it's not an important field
                entrypoint = section[i].VirtualAddress + vxdhdr->e32_eip;
            }
        }
    } else {
        fileasmsz   = Decode16Bits;
        imagebase   = 0;
        sec_align   = 0;
        entrypoint  = imagebase + (doshdr->e_cs < 0x8000) ? doshdr->e_ip : 0;
        sections    = 0;
    }
    return(p - filemem);
}



int parse_ELF32(HWND hwnd) {
    elf32_header_t  *elfhdr;
    elf32_section_t *elfsec;
    int     i;
    u8      *p;

    p = filemem;
    elfhdr = (elf32_header_t *)p;     p += sizeof(elf32_header_t);
    if(memcmp(elfhdr->e_ident, "\x7f""ELF", 4)) return(-1);
    if(elfhdr->e_ident[4] != 1) return(-1); // only 32 bit supported
    if(elfhdr->e_ident[5] != 1) return(-1); // only little endian

    fileasmsz   = Decode32Bits;
    imagebase   = 0;
    sec_align   = 0;
    entrypoint  = elfhdr->e_entry;

    sections = elfhdr->e_shnum;
    section = calloc(sizeof(section_t), sections);
    if(!section) std_err(hwnd);

    elfsec = (elf32_section_t *)(filemem + elfhdr->e_shoff);
    for(i = 0; i < sections; i++) {
        strncpy(section[i].Name, filemem + elfsec[elfhdr->e_shstrndx].sh_offset + elfsec[i].sh_name, SECNAMESZ);
        section[i].Name[SECNAMESZ]  = 0;
        section[i].VirtualAddress   = elfsec[i].sh_addr;
        section[i].VirtualSize      = elfsec[i].sh_size;
        section[i].PointerToRawData = elfsec[i].sh_offset;
        section[i].SizeOfRawData    = elfsec[i].sh_size;
        section[i].Characteristics  = elfsec[i].sh_flags;
        if(!section[i].VirtualSize) section[i].VirtualSize = section[i].SizeOfRawData;  // Watcom
    }
    return(p - filemem);
}



int parse_ELF64(HWND hwnd) {
    elf64_header_t  *elfhdr;
    elf64_section_t *elfsec;
    int     i;
    u8      *p;

    p = filemem;
    elfhdr = (elf64_header_t *)p;     p += sizeof(elf64_header_t);
    if(memcmp(elfhdr->e_ident, "\x7f""ELF", 4)) return(-1);
    if(elfhdr->e_ident[4] != 2) return(-1); // only 64 bit supported
    if(elfhdr->e_ident[5] != 1) return(-1); // only little endian

    fileasmsz   = Decode64Bits;
    imagebase   = 0;
    sec_align   = 0;
    entrypoint  = elfhdr->e_entry;

    sections = elfhdr->e_shnum;
    section = calloc(sizeof(section_t), sections);
    if(!section) std_err(hwnd);

    elfsec = (elf64_section_t *)(filemem + elfhdr->e_shoff);
    for(i = 0; i < sections; i++) {
        strncpy(section[i].Name, filemem + elfsec[elfhdr->e_shstrndx].sh_offset + elfsec[i].sh_name, SECNAMESZ);
        section[i].Name[SECNAMESZ]  = 0;
        section[i].VirtualAddress   = elfsec[i].sh_addr;
        section[i].VirtualSize      = elfsec[i].sh_size;
        section[i].PointerToRawData = elfsec[i].sh_offset;
        section[i].SizeOfRawData    = elfsec[i].sh_size;
        section[i].Characteristics  = elfsec[i].sh_flags;
        if(!section[i].VirtualSize) section[i].VirtualSize = section[i].SizeOfRawData;  // Watcom
    }
    return(p - filemem);
}



int parse_XBE(HWND hwnd) {
    xbe_header_t    *xbe_header;
    xbe_section_t   *xbe_section;
    int     i;
    u8      *p;

    p = filemem;
    xbe_header = (xbe_header_t *)p;     p += sizeof(xbe_header_t);
    if(xbe_header->magic_id != 0x48454258) return(-1);  // XBEH

    fileasmsz   = Decode32Bits;
    imagebase   = xbe_header->base_address;
    sec_align   = MYALIGNMENT;  // ???
    entrypoint  = imagebase + xbe_header->entry_point;
    sections    = xbe_header->number_of_sections;

    section = calloc(sizeof(section_t), sections);
    if(!section) std_err(hwnd);

    xbe_section = (xbe_section_t *)(filemem + xbe_header->section_header_address - imagebase);
    for(i = 0; i < sections; i++) {
        strncpy(section[i].Name, filemem + xbe_section[i].section_name_address - imagebase, SECNAMESZ);
        section[i].Name[SECNAMESZ]  = 0;
        section[i].VirtualAddress   = xbe_section[i].virtual_address;
        section[i].VirtualSize      = xbe_section[i].virtual_size;
        section[i].PointerToRawData = xbe_section[i].raw_address;
        section[i].SizeOfRawData    = xbe_section[i].raw_size;
        section[i].Characteristics  = xbe_section[i].section_flags;
    }
    return(p - filemem);
}



int parse_file(HWND hwnd) {
    int     offset;

                   offset = parse_PE(hwnd);
    if(offset < 0) offset = parse_ELF32(hwnd);
    if(offset < 0) offset = parse_ELF64(hwnd);
    if(offset < 0) offset = parse_XBE(hwnd);
    if(offset < 0) {
        global_reset(hwnd);
        MessageBox(hwnd, "this is not a valid DOS/PE/ELF/XBE file or isn't yet supported by this program", VER, MB_ICONERROR | MB_OK | MB_TASKMODAL);
        return(-1);
    }
    if(!sections || !section) { // possible work-around in case of errors
        sections = 1;
        if(section) free(section);
        section = calloc(sizeof(section_t), sections);
        if(!section) std_err(hwnd);
        section[0].VirtualAddress   = 0;
        section[0].VirtualSize      = filememsz - offset;
        section[0].PointerToRawData = offset;
        section[0].SizeOfRawData    = filememsz - offset;
        section[0].Characteristics  = 0;
    }
    SetDlgItemText(hwnd, IDC_IMAGEBASE,  itox(imagebase));
    SetDlgItemText(hwnd, IDC_SECALIGN,   itox(sec_align));
    SetDlgItemText(hwnd, IDC_ENTRYPOINT, itox(entrypoint));
    return(0);
}



void load_process(HWND hwnd, DWORD pid, BYTE *baddr, DWORD size) {
    HANDLE  process;
    u8      question[100];

    process = OpenProcess(PROCESS_VM_READ, FALSE, pid);
    if(!process) {
        winerr(hwnd);
        return;
    }

    global_reset(hwnd);

    if(size > 16777216) {
        sprintf(question, "this process is bigger than 16 megabytes: %u bytes\n" "do you want to load all?", (u32)size);
        if(MessageBox(hwnd, question, VER, MB_YESNO | MB_TASKMODAL) == IDNO) {
            size = 16777216;
        }
    }

    filemem = malloc(size);
    if(!filemem) {
        CloseHandle(process);
        std_err(hwnd);
        return;
    }

    filememsz = size;
    if(!ReadProcessMemory(
      process,
      (LPCVOID)baddr,
      filemem,
      filememsz, NULL)) {
        CloseHandle(process);
        winerr(hwnd);
        free(filemem);
        filemem = 0;
        return;
    }

    CloseHandle(process);

    if(parse_file(hwnd) < 0) return;

    isprocess = 1;
    SetDlgItemText(hwnd, IDC_FILENAME,  filename);
    SetDlgItemText(hwnd, IDC_FILESIZE,  itox(filememsz));
    if(set_offset(hwnd, entrypoint, IDC_RVA) < 0) set_offset(hwnd, 0, IDC_FILE);
}



void load_file(HWND hwnd) {
    struct stat xstat;
    FILE        *fd;
    int         len;
    u8          question[100];

    if(filename[0] == '\"') {
        len = strlen(filename) - 1;
        filename[len] = 0;
        memmove(filename, filename + 1, len);
    }

    fd = fopen(filename, "rb");
    if(!fd) {
        std_err(hwnd);
        return;
    }

    global_reset(hwnd);

    fstat(fileno(fd), &xstat);

    if(xstat.st_size > 16777216) {
        sprintf(question, "this file is bigger than 16 megabytes: %u bytes\n" "do you want to load all?", (u32)xstat.st_size);
        if(MessageBox(hwnd, question, VER, MB_YESNO | MB_TASKMODAL) == IDNO) {
            xstat.st_size = 16777216;
        }
    }

    filemem = malloc(xstat.st_size);
    if(!filemem) {
        fclose(fd);
        std_err(hwnd);
        return;
    }
    filememsz = fread(filemem, 1, xstat.st_size, fd);
    fclose(fd);

    if(parse_file(hwnd) < 0) return;

    SetDlgItemText(hwnd, IDC_FILENAME,  filename);
    SetDlgItemText(hwnd, IDC_FILESIZE,  itox(filememsz));
    if(set_offset(hwnd, entrypoint, IDC_RVA) < 0) set_offset(hwnd, 0, IDC_FILE);
}



void get_file(HWND hwnd) {
    OPENFILENAME    ofn;
    static const u8 filter[] =
                    "PE/ELF files (*.exe;*.dll;*.ocx;*.cpl;*.scr;*.out)\0" "*.exe;*.dll;*.ocx;*.cpl;*.scr;*.out\0"
                    "(*.*)\0"   "*.*\0"
                    "\0\0";

    filename[0] = 0;
    memset(&ofn, 0, sizeof(ofn));
    ofn.lStructSize     = sizeof(ofn);
    ofn.hwndOwner       = hwnd;
    ofn.lpstrInitialDir = ".";
    ofn.lpstrFilter     = filter;
    ofn.nFilterIndex    = 1;
    ofn.lpstrFile       = filename;
    ofn.nMaxFile        = sizeof(filename);
    ofn.lpstrTitle      = "Select the PE or ELF file";
    ofn.Flags           = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_LONGNAMES | OFN_EXPLORER | OFN_READONLY;

    if(!GetOpenFileName(&ofn)) return;

    load_file(hwnd);
}



u32 search_string(u32 offset, u8 *str, int direction) {
    int     n,
            len;
    u8      cstr[SEARCHSZ],
            *limit,
            *p,
            *o;

    if(!filemem) return(-1);
    p = str;
    o = cstr;
    while(*p) {
        if(*p == '$') {
            if(sscanf(p + 1, "%02x%n", &n, &len) != 1) return(-1);
            p += len;
            *o++ = n;
        } else if(*p == '\\') {
            p++;
            switch(*p) {
                case 0:  return(-1); break;
                case '0':  n = '\0'; break;
                case 'a':  n = '\a'; break;
                case 'b':  n = '\b'; break;
                case 'e':  n = '\e'; break;
                case 'f':  n = '\f'; break;
                case 'n':  n = '\n'; break;
                case 'r':  n = '\r'; break;
                case 't':  n = '\t'; break;
                case 'v':  n = '\v'; break;
                case '\"': n = '\"'; break;
                case '\'': n = '\''; break;
                case '\\': n = '\\'; break;
                case '?':  n = '\?'; break;
                case '.':  n = '.';  break;
                case 'x': {
                    if(sscanf(p + 1, "%02x%n", &n, &len) != 1) return(-1);
                    p += len;
                    } break;
                default: {
                    if(sscanf(p + 1, "%o%n", &n, &len) != 1) return(-1);
                    p += len;
                    } break;
            }
            *o++ = n;
        } else {
            *o++ = *p;
        }
        p++;
    }
    len = o - cstr;
    if(!len) return(-1);

    if(direction == VK_NEXT) {
        limit = filemem + filememsz - len;
        for(p = filemem + offset; p <= limit; p++) {
            if(!memcmp(p, cstr, len)) return(p - filemem);
        }
    } else {
        for(p = filemem + offset; p >= filemem; p--) {
            if(!memcmp(p, cstr, len)) return(p - filemem);
        }
    }
    return(-1);
}



u32 get_offset(HWND hwnd, int dlg) {
    u32     offset;
    u8      buff[14],
            *p;

    GetDlgItemText(hwnd, dlg, buff, sizeof(buff));
    if(!buff[0]) strcpy(buff, "0");
    p = strchr(buff, 'x');
    if(!p) p = strchr(buff, 'X');
    if(!p) {
        p = buff;
    } else {
        p++;
    }
    if(sscanf(p, "%x", &offset) != 1) return(0);
    return(offset);
}



BOOL CALLBACK PrcProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam) {
    HFONT   hf;
    HWND    hlist;
    DWORD   pid,
            size;
    int     sel;
    BYTE    *baddr;
    u8      buff[TEXTSZ];

    switch(Message) {
        case WM_INITDIALOG: {
            hf = CreateFont(10, 0, 0, 0, 0, FALSE, 0, 0, 0, 0, 0, 0, 0, "Lucida Console");
            SendDlgItemMessage(hwnd, IDC_PROCLIST, WM_SETFONT, (WPARAM)hf, TRUE);
            process_list(hwnd);
            } break;
        case WM_COMMAND: {
            switch(LOWORD(wParam)) {
                case IDC_PROCLIST: {
                    if(HIWORD(wParam) == LBN_DBLCLK) {
                        SendMessage(hwnd, WM_COMMAND, (WPARAM)IDOK, (LPARAM)0);
                    }
                    } break;
                case IDOK: {
                    hlist = GetDlgItem(hwnd, IDC_PROCLIST);
                    sel = SendMessage(hlist, LB_GETCURSEL, (WPARAM)0, (LPARAM)0);
                    if(sel < 0) break;
                    SendMessage(hlist, LB_GETTEXT, (WPARAM)sel, (LPARAM)buff);
                    if((buff[0] == '-') || (buff[7] == 'p')) break;
                    sscanf(buff, PROCFMT, (u32 *)(void *)&pid, (u32 *)(void *)&baddr, (u32 *)(void *)&size, filename);
                    sprintf(filename + strlen(filename), " (process %u)", (u32)pid);
                    load_process(GetNextWindow(hwnd, GW_HWNDNEXT), pid, baddr, size);
                    DestroyWindow(hwnd);
                    } break;
                case IDCANCEL: {
                    DestroyWindow(hwnd);
                    } break;
                default: {
                    } break;
            }
            } break;
        case WM_CLOSE: {
            DestroyWindow(hwnd);
            } break;
        case WM_DESTROY: {
            } break;
        default: {
            return(0);
            } break;
    }
    return(0);
}



int check_key_pressed(int key) {
    static  u32 oldkey  = 0,
                oldtime = 0;

    if((GetAsyncKeyState(key) & 0x8000) && ((oldkey != key) || ((GetTickCount() - oldtime) > 100))) {
        oldkey  = key;
        oldtime = GetTickCount();
        return(TRUE);
    }
    return(FALSE);
}



BOOL CALLBACK DlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam) {
    HFONT   hf;
    u32     offset;
    int     idx;
    u8      search[SEARCHSZ];

    switch(Message) {
        case WM_INITDIALOG: {
            //ideal = 0; lowercase = 0; tabarguments = 1; extraspace = 1; putdefseg = 0; showmemsize = 1;
            //shownear = 0; shortstringcmds = 0; symbolic = 1; farcalls = 1; decodevxd = 1; privileged = 1;
            //iocommand = 1; badshift = 1; extraprefix = 1; lockedbus = 1; stackalign = 1; iswindowsnt = 0;

            global_reset(hwnd);
            DragAcceptFiles(hwnd, TRUE);
            hf = CreateFont(10, 0, 0, 0, 0, FALSE, 0, 0, 0, 0, 0, 0, 0, "Lucida Console");
            SendDlgItemMessage(hwnd, IDC_BYTES,      WM_SETFONT, (WPARAM)hf, TRUE);
            SendDlgItemMessage(hwnd, IDC_DISASM,     WM_SETFONT, (WPARAM)hf, TRUE);

            SetTimer(hwnd, IDC_TIMER, 20, (TIMERPROC)NULL);
            if(commandline && commandline[0]) {
                sprintf(filename, "%.*s", sizeof(filename), commandline);
                load_file(hwnd);
            }
            } break;
        case WM_COMMAND: {
            switch(LOWORD(wParam)) {
                case IDC_OPEN_FILE: {
                    get_file(hwnd);
                    } break;
                case IDC_OPEN_PROC: {
                    return(DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_PROC), NULL, PrcProc));
                    } break;
                case IDC_SEARCH:
                case IDC_SEARCHPREV:
                case IDC_SEARCHNEXT: {
                    if(!filemem) break;
                    GetDlgItemText(hwnd, IDC_SEARCH, search, sizeof(search));
                    if(!search[0]) break;
                    offset = curroff;
                    if(LOWORD(wParam) == IDC_SEARCHPREV) offset--;
                    if(LOWORD(wParam) == IDC_SEARCHNEXT) offset++;
                    offset = search_string(
                        offset,
                        search,
                        (LOWORD(wParam) == IDC_SEARCHPREV) ? VK_PRIOR : VK_NEXT);
                    if(offset == -1) break;
                    set_offset(hwnd, offset, IDC_FILE);
                    } break;
                case IDC_ABOUT: {
                    MessageBox(hwnd, VER2, VER, MB_OK | MB_TASKMODAL);
                    } break;
                default: {
                    } break;
            }
            } break;
        case WM_TIMER: {
            if(GetForegroundWindow() != hwnd) break;
            if(check_key_pressed(VK_CONTROL)) {
                if(check_key_pressed('G')) SetFocus(GetDlgItem(hwnd, IDC_FILE));
                else if(check_key_pressed('F')) SetFocus(GetDlgItem(hwnd, IDC_SEARCH));
                else if(check_key_pressed('O')) SendMessage(hwnd, WM_COMMAND, (WPARAM)IDC_OPEN_FILE, (LPARAM)0);
                else if(check_key_pressed('P')) SendMessage(hwnd, WM_COMMAND, (WPARAM)IDC_OPEN_PROC, (LPARAM)0);
                else if(check_key_pressed('B')) SetFocus(GetDlgItem(hwnd, IDC_BYTES));
                else if(check_key_pressed('D')) SetFocus(GetDlgItem(hwnd, IDC_DISASM));
                else if(check_key_pressed('Q')) SendMessage(hwnd, WM_CLOSE, (WPARAM)0, (LPARAM)0);
            } else if(check_key_pressed(VK_SHIFT)) {
                if(check_key_pressed(VK_F3)) SendMessage(hwnd, WM_COMMAND, (WPARAM)IDC_SEARCHPREV, (LPARAM)0);
            } else {
                if(check_key_pressed(VK_F1)) SendMessage(hwnd, WM_COMMAND, (WPARAM)IDC_ABOUT, (LPARAM)0);
                else if(check_key_pressed(VK_F3)) SendMessage(hwnd, WM_COMMAND, (WPARAM)IDC_SEARCHNEXT, (LPARAM)0);
                else if(check_key_pressed(VK_UP)) {
                    if(GetFocus() == GetDlgItem(hwnd, IDC_BYTES)) set_offset(hwnd, curroff - 16, IDC_FILE);
                    else if(GetFocus() == GetDlgItem(hwnd, IDC_DISASM)) dumpasm(hwnd, -1);
                    else if(GetFocus() == GetDlgItem(hwnd, IDC_NAME)) {
                        for(idx = get_section(curroff) + 1; idx < sections; idx++) {    // simple way
                            if(section[idx].PointerToRawData > curroff) {
                                set_offset(hwnd, section[idx].PointerToRawData, IDC_FILE);
                                break;
                            }
                        }
                    } 
                } else if(check_key_pressed(VK_LEFT)) {
                    if(GetFocus() == GetDlgItem(hwnd, IDC_BYTES)) set_offset(hwnd, curroff - 1, IDC_FILE);
                    else if(GetFocus() == GetDlgItem(hwnd, IDC_DISASM)) set_offset(hwnd, curroff - 1, IDC_FILE);
                } else if(check_key_pressed(VK_RIGHT)) {
                    if(GetFocus() == GetDlgItem(hwnd, IDC_BYTES)) set_offset(hwnd, curroff + 1, IDC_FILE);
                    else if(GetFocus() == GetDlgItem(hwnd, IDC_DISASM)) set_offset(hwnd, curroff + 1, IDC_FILE);
                } else if(check_key_pressed(VK_DOWN)) {
                    if(GetFocus() == GetDlgItem(hwnd, IDC_BYTES)) set_offset(hwnd, curroff + 16, IDC_FILE);
                    else if(GetFocus() == GetDlgItem(hwnd, IDC_DISASM)) dumpasm(hwnd, 1);
                    else if(GetFocus() == GetDlgItem(hwnd, IDC_NAME)) {
                        for(idx = get_section(curroff) - 1; idx >= 0; idx--) {    // simple way
                            if(section[idx].PointerToRawData < curroff) {
                                set_offset(hwnd, section[idx].PointerToRawData, IDC_FILE);
                                break;
                            }
                        }
                    }
                } else if(check_key_pressed(VK_PRIOR)) {
                    if(GetFocus() == GetDlgItem(hwnd, IDC_BYTES)) set_offset(hwnd, curroff - MAXPAGE, IDC_FILE);
                    else if(GetFocus() == GetDlgItem(hwnd, IDC_DISASM)) dumpasm(hwnd, -MAXLINES);
                } else if(check_key_pressed(VK_NEXT)) {
                    if(GetFocus() == GetDlgItem(hwnd, IDC_BYTES)) set_offset(hwnd, curroff + MAXPAGE, IDC_FILE);
                    else if(GetFocus() == GetDlgItem(hwnd, IDC_DISASM)) dumpasm(hwnd, MAXLINES);
                } else if(check_key_pressed(VK_RETURN)) {
                    if(GetFocus() == GetDlgItem(hwnd, IDC_RVA)) {
                        offset = get_offset(hwnd, IDC_RVA);
                        set_offset(hwnd, offset,  IDC_RVA);
                    } else if(GetFocus() == GetDlgItem(hwnd, IDC_FILE)) {
                        offset = get_offset(hwnd, IDC_FILE);
                        set_offset(hwnd, offset,  IDC_FILE);
                    }
                }
            }
            } break;
        case WM_DROPFILES: {
            DragQueryFile((HDROP)wParam, 0, filename, sizeof(filename));
            DragFinish((HDROP)wParam);
            load_file(hwnd);
            } break;
        case WM_CLOSE: {
            KillTimer(hwnd, IDC_TIMER);
            DestroyWindow(hwnd);
            } break;
        case WM_DESTROY: {
            PostQuitMessage(0);
            } break;
        default: {
            return(0);
            } break;
    }
    return(0);
}



int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd) {
    commandline = lpCmdLine;
    return(DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, DlgProc));
}


